library ac_dart.auth;

import 'dart:io' hide HttpException;
import 'dart:async';

import 'package:option/option.dart';
import 'package:either/either.dart';
import 'package:atlassian_connect_host/atlassian_connect_host.dart';

import 'package:logging/logging.dart';
import 'package:shelf/shelf.dart';
import 'package:shelf_auth/shelf_auth.dart';
export 'package:shelf/shelf.dart' show Request, Response, Handler, Middleware;
import 'package:http_exception/http_exception.dart';
import 'package:atlassian_connect_shelf/router.dart';

Logger _log = new Logger('ac_dart.auth');

const AC_DART_SESSION_PARAM = 'atlassian-acdart-session';
@deprecated
const AC_DART_SESSION_PARAM_HOST_KEY = '$AC_DART_SESSION_PARAM-hostKey';
@deprecated
const AC_DART_SESSION_PARAM_HOST_URL = '$AC_DART_SESSION_PARAM-hostUrl';
@deprecated
const AC_DART_SESSION_PARAM_USERNAME = '$AC_DART_SESSION_PARAM-userName';
@deprecated
const AC_DART_SESSION_PARAM_TOKEN = '$AC_DART_SESSION_PARAM-token';
@deprecated
const AC_DART_SESSION_RESPONSE_HEADER = 'X-$AC_DART_SESSION_PARAM_TOKEN';

//abstract class Authenticator {
//  dynamic authenticate(Request request, Handler innerHandler);
//
//  Middleware get middleware;
//}

const String AC_DART_CONTEXT_PARAM_REQUEST_CONTEXT =
    'AC_DART_CONTEXT_PARAM_REQUEST_CONTEXT';

Option<AcDartRequestContext> getRequestContext(Request request) =>
    new Option(request.context[AC_DART_CONTEXT_PARAM_REQUEST_CONTEXT]);

Request setRequestContext(
        Request request, AcDartRequestContext requestContext) =>
    request.change(
        context: {AC_DART_CONTEXT_PARAM_REQUEST_CONTEXT: requestContext});

abstract class AbstractHostServiceAuthenticator
    extends AbstractAuthenticator<AcPrincipal> {
  final String queryParamName;

  AbstractHostServiceAuthenticator(this.queryParamName) : super(true, true);

  @override
  Future<Option<AuthenticatedContext<AcPrincipal>>> authenticate(
      Request request) {
    _log.finer('authenticating ${request.method} ${request.requestedUri}');

    final Option<String> jwtTokenOpt =
        new Option(request.requestedUri.queryParameters[queryParamName]).orElse(
            () => _fromAuthHeader(request.headers[HttpHeaders.AUTHORIZATION]));

    return jwtTokenOpt.map((jwtToken) async {
      _log.finest('token is $jwtToken');

      final requestContextOpt = getRequestContext(request);
      if (requestContextOpt is None) {
        throw new HttpException(500, 'missing host service in request context');
      }

      final Either authResult = await authenticateRequest(
          jwtToken, request, requestContextOpt.get().hostService);
      if (authResult.isRight()) {
        _log.finer('request successfully authenticated ${request.method} '
            '${request.requestedUri}');
        final AcPrincipal principal = authResult.right().get();
        // TODO: This should be a SessionAuthenticatedContext!!!!!
        // OR does this happen in the session handler
        return new Some(createContext(principal));
      } else {
        _log.finer('request failed authentication ${request.method} '
            '${request.requestedUri}');
        AuthenticationFailure authenticationFailure = authResult.left().get();
        throw authenticationFailure.statusCode == HttpStatus.UNAUTHORIZED
            ? new UnauthorizedException(
                {"error": authenticationFailure.message},
                authenticationFailure.message)
            : new ForbiddenException({"error": authenticationFailure.message},
                authenticationFailure.message);
      }
    }).getOrElse(() {
      _log.finer('request failed authentication - no jwt in request: '
          '${request.method} ${request.requestedUri}');
      return new Future.value(const None());
//      return new Response(HttpStatus.UNAUTHORIZED,
//          body: 'no jwt token found in request');
    });
  }

  Option<String> _fromAuthHeader(String authHeader) {
    List<String> authHeaders = authHeader == null ? [] : authHeader.split(',');
    final firstJwtHeader = authHeaders.firstWhere(
        (authHeader) => authHeader.startsWith('JWT'),
        orElse: () => '');

    final authTypeCredentialPair = firstJwtHeader.split(' ');
    if (authTypeCredentialPair.length != 2) {
      return const None();
    }

    return new Some(authTypeCredentialPair[1]);
  }

  Future<Either<AuthenticationFailure, AcPrincipal>> authenticateRequest(
      String jwtToken, Request request, AcHostService hostService);
}

class JwtAuthenticator extends AbstractHostServiceAuthenticator {
  JwtAuthenticator() : super('jwt');

  @override
  Future<Either<AuthenticationFailure, AcPrincipal>> authenticateRequest(
      String jwtToken, Request request, AcHostService hostService) async {
    return hostService.authenticateHostRequest(
        jwtToken, request.requestedUri, request.method);
  }
}

class SessionTokenAuthenticator extends AbstractAuthenticator<AcPrincipal> {
  SessionTokenAuthenticator() : super(false, true);

  @override
  Future<Option<AuthenticatedContext<AcPrincipal>>> authenticate(
      Request request) {
    final requestContextOpt = getRequestContext(request);
    if (requestContextOpt is None) {
      throw new HttpException(500, 'missing host service in request context');
    }
    return requestContextOpt
        .get()
        .hostService
        .jwtSessionHandler
        .authenticator
        .authenticate(request);
  }
}

class SessionTokenSessionHandler extends SessionHandler<AcPrincipal> {
  @override
  Authenticator<AcPrincipal> get authenticator =>
      new SessionTokenAuthenticator();

  @override
  Future<Response> handle(
      AuthenticatedContext context, Request request, Response response) {
    final requestContextOpt = getRequestContext(request);
    if (requestContextOpt is None) {
      throw new HttpException(500, 'missing host service in request context');
    }
    return requestContextOpt
        .get()
        .hostService
        .jwtSessionHandler
        .handle(context, request, response);
  }
}

AcPrincipal getPrincipal(Request request) {
  return getAuthenticatedContext(request)
      .map((AuthenticatedContext c) => c.principal)
      .getOrElse(() => null);
}
