library ac_dart.addon_routebuilder.common.impl;

import 'package:shelf/shelf.dart';
import 'package:uri/uri.dart';
import 'package:mojito/mojito.dart' hide Router, router;
import 'package:atlassian_connect_shelf/acdart_shelf_middleware.dart';
import 'package:atlassian_connect_shelf/router.dart';

import 'package:path/path.dart' as p;
import 'package:logging/logging.dart';
import 'package:atlassian_connect_shelf/src/router/common/common_routes.dart';
import 'package:atlassian_connect_shelf/src/descriptor_builders/descriptor_builder.dart';
import 'package:atlassian_connect_host/atlassian_connect_host.dart';
import 'package:http_exception/http_exception.dart';
import 'dart:convert';
import 'package:atlassian_connect_shelf/src/auth.dart';
import 'package:option/option.dart';

final Logger _log = new Logger('ac_dart.addon_routebuilder.common');

abstract class CommonAddonRouterBuilderImpl<B extends DescriptorBuilder,
    S extends Scope> implements CommonAddonRouterBuilder {
  final Router router;
  final B descriptorBuilder;
  final AcDartRequestContext requestContext;
  final Middleware _standardAuthMiddleware;
  final Option<ClientIdFactory> _oauthClientIdFactoryOpt;

  static HandlerAdapter _createHandlerAdapter(
      AcDartRequestContext requestContext,
      {Map<Type, CustomObjectFactory> customObjects: const {}}) {
    final Map<Type, CustomObjectFactory> _defaultCustomObjects = {
      AcDartRequestContext: (request) => requestContext,
      AcHostService: (_) => requestContext.hostService,
      AcPrincipal: getPrincipal,
      AcHostHttpClient: (req) => requestContext.httpClient,
      AcHostRegistration: (req) => requestContext.hostRegistration
    };

    final _customObjects = <Type, CustomObjectFactory>{};
    _customObjects.addAll(_defaultCustomObjects);

    if (customObjects != null) {
      _customObjects.addAll(customObjects);
    }

    final _handlerAdapter = handlerAdapter(
        customObjects: _customObjects,
        // false till https://bitbucket.org/andersmholmgren/constraint/issue/34/constraint-resolver-does-not-cope-with-all
        validateParameters: false);
    return _handlerAdapter;
  }

  static Middleware _addContext(AcDartRequestContext requestContext) =>
      (Handler innerHandler) => (Request request) =>
          innerHandler(setRequestContext(request, requestContext));

  static Middleware _createSpaAuthMiddleware(SessionHandler sessionHandler,
      AcHostService hostService, IsDevMode isDevMode) {
    final _sessionHandler =
        sessionHandler != null ? sessionHandler : hostService.jwtSessionHandler;

    // TODO: could in theory pass hostService to JwtAuthenticator ctr like we
    // used to
    return (new AuthenticationBuilder()
        .authenticator(new JwtAuthenticator())
        .sessionHandler(_sessionHandler)
          ..allowAnonymousAccess = false
          ..allowHttp = isDevMode()).build();
  }

  CommonAddonRouterBuilderImpl._(
      Router router,
      path,
      this.descriptorBuilder,
      this._oauthClientIdFactoryOpt,
      AcDartRequestContext requestContext,
      SessionHandler sessionHandler,
      IsDevMode isDevMode)
      : this.router = router.child(path,
            handlerAdapter: _createHandlerAdapter(requestContext),
            middleware: _addContext(requestContext)),
        this.requestContext = requestContext,
        this._standardAuthMiddleware = _createSpaAuthMiddleware(
            sessionHandler, requestContext.hostService, isDevMode);

  CommonAddonRouterBuilderImpl(
      Router router,
      path,
      B descriptorBuilder,
      String addonKey,
      Uri baseUrl,
      AcHostRegistrationRepository hostRegistrationRepository,
      IsDevMode isDevMode,
      {SessionHandler sessionHandler,
      String addonSecret,
      ClientIdFactory oauthClientIdFactory})
      : this._(
            router,
            path,
            descriptorBuilder,
            new Option(oauthClientIdFactory),
            new AcDartRequestContext(new AcHostService(
                hostRegistrationRepository,
                new AddonHost(addonKey, baseUrl, addonSecret))),
            sessionHandler,
            isDevMode);

  void addInstallationRoute({String path: 'installed'}) {
    router.post(path, defaultInstallationHandler,
        name: DescriptorBuilder.installationRouteName);
  }

  @override
  void addDescriptorRoute(Vendor vendor, String name, String description,
      {String path: 'descriptor', Iterable<S> scopes: const []}) {
    Map json;

    router.get(path, (Request request) async {
      final routingContext = getRoutingContext(request).get();

      final oauthClientId = await _oauthClientIdFactoryOpt
          .map((factory) => factory(request))
          .getOrDefault(null);

      if (json == null) {
        json = descriptorBuilder
            .build(routingContext.routingOrder.first, vendor,
                name: name,
                description: description,
                oauthClientId: oauthClientId,
                scopes: scopes)
            .toJson();
      }

      return json;
    });
  }

  void addSpaRedirect(path, String name,
      {String spaBasePath: 'ui',
      String fragmentPath,
      Map<String, String> additionalUriFragmentParams: const {}}) {
    router.add(
        path,
        ['GET'],
        _convertToFragment(
            spaBasePath, fragmentPath, additionalUriFragmentParams),
        exactMatch: false,
        name: name,
        middleware: new MiddlewareBuilder()
            .addMiddleware(_moveResponseTokenToFragment())
            .addMiddleware(_standardAuthMiddleware)
            .build());
  }

  Handler _convertToFragment(String spaBasePath, String fragmentPath,
      Map<String, String> additionalUriFragmentParams) {
    return (Request request) {
      // http://foo.com/sparedirect/app.html?jwt=foo&user=blah
      // => http://foo.com/ui/app.html#?sessionToken=myToken&user=fred
      // probably won't work with auth header as the header will be lost before
      // the app loads.
      // This result not cacheable => ensure!!
      // but the app is
      //

      final builder = new UriBuilder.fromUri(request.requestedUri);
      builder.path = p.join(
          spaBasePath, p.basename(request.handlerPath), request.url.path);
      final queryParameters = builder.queryParameters;
      queryParameters.remove('jwt');
      queryParameters.addAll((additionalUriFragmentParams));
      builder.queryParameters = const {};
      String _fragmentPath = fragmentPath != null ? fragmentPath : '';
      builder.fragment =
          '$_fragmentPath?${_encodeQueryParams(queryParameters)}';

      final redirectUri = builder.build();
      _log.finest('redirecting from ${request.requestedUri} to $redirectUri');
      return new Response.movedPermanently(redirectUri);
    };
  }

  // note: this isn't pretty but there is no way to access the headers for a page
  // within the browser. A common alternative is to use meta tags but here
  // we aren't returning html, just a redirect. So the best place is a uri
  // fragment. Can't really push this to shelf_auth as the redirecting strategy
  // is more an addon style thing I think.
  Middleware _moveResponseTokenToFragment() {
    return createMiddleware(responseHandler: (Response response) {
      if (response.statusCode > 399) {
        return response;
      }
      final AuthorizationHeader authHeader =
          responseAuthorizationHeader(response, JWT_SESSION_AUTH_SCHEME)
              .getOrElse(() => throw new StateError('missing auth header'));

      final builder =
          new UriBuilder.fromUri(Uri.parse(response.headers['location']));
      final params = Uri.splitQueryString(builder.fragment);
      params['sessionToken'] = authHeader.credentials;
      builder.fragment = _encodeQueryParams(params);
      return response.change(headers: {'location': builder.toString()});
    });
  }

  // Ouch
  String _encodeQueryParams(Map<String, String> queryParams) {
    final builder = new UriBuilder()..queryParameters = queryParams;
    return builder.build().query;
  }
}

Future defaultInstallationHandler(
    InstallationEvent installationEvent, AcHostService hostService) async {
  _log.fine('received installation request '
      '${_prettyJson.convert(installationEvent)}');

  try {
    await hostService.registerHost(installationEvent);

    return "installed";
  } catch (e, stackTrace) {
    _log.warning('Registration failed', e, stackTrace);
    // TODO: more specific code and message. e.g 404, 401, 403
    throw new BadRequestException({}, 'Regstration Failed');
  }
}

final JsonEncoder _prettyJson = new JsonEncoder.withIndent('  ');
