library ac_dart.auth.test;

import 'package:atlassian_connect_shelf/src/auth.dart';
import 'package:test/test.dart';
import 'package:mockito/mockito.dart';
import 'dart:async';

import 'package:atlassian_connect_jwt/atlassian_connect_jwt.dart';
import 'package:atlassian_connect_host/atlassian_connect_host.dart';
import 'package:atlassian_connect_host/host_model.dart';
import 'package:logging/logging.dart';
import 'package:either/either.dart';
import 'package:option/option.dart';
import 'package:mojito/mojito.dart'
    show Authenticator, MiddlewareBuilder, authenticate;
import 'package:atlassian_connect_shelf/acdart_shelf_middleware.dart';
import 'package:shelf_exception_handler/shelf_exception_handler.dart';
import 'package:atlassian_connect_shelf/router.dart';

class _MockAcHostService extends Mock implements AcHostService {
  noSuchMethod(i) => super.noSuchMethod(i);
}

//mocking: typedef Handler(Request request);
abstract class HandlerClass {
  dynamic call(Request request) {
    return null;
  }
}

class MockHandler extends Mock implements HandlerClass {
  noSuchMethod(Invocation invocation) => super.noSuchMethod(invocation);
}

void _initLogging() {
  Logger.root.level = Level.WARNING;
  Logger.root.onRecord.listen((LogRecord rec) {
    print('${rec.message}');
  });
}

void main() {
  _initLogging();

  MockHandler handler;

  _MockAcHostService hostService;

  Request request;

  Middleware authenticator;

  final String hostKey = "jira:ae390d29-31b2-4c12-a719-9df64e3e92b7";
  final String user = 'fred';
  final String sharedSecret = '3ab90b11-d7bd-4097-958f-01b7ac4e985f';
  final String sessionToken = 'la tokena';

  final hostRego = (new AcHostRegistration.build(
      sharedSecret: "$sharedSecret",
      hostDetails: (new AcHostDetails.build(
          key: "$hostKey",
          baseUrl: Uri.parse('http://localhost:8080/'),
          productType: ProductType.jira,
          publicKey: "pubby"))));

  setUp(() {
    hostService = new _MockAcHostService();
    handler = new MockHandler();
    when(handler.call(any)).thenReturn(new Response.ok("woohoo"));

    when(hostService.findRegistrationByHostKey(any))
        .thenReturn(new Future.value(new Some(hostRego)));

    when(hostService.createAddonSessionToken(any, any))
        .thenReturn(sessionToken);

//    when(hostService.jwtSessionHandler).thenReturn();
  });

  void mockRequest(Uri uri) {
    request = setRequestContext(
        new Request("GET", uri), new AcDartRequestContext(hostService));
  }

  Future<Response> authenticateRequest(Uri uri) {
    mockRequest(uri);
    return new Future.value(authenticator(handler)(request));
  }

  _expect(Uri uri, int expectedResponseStatus) async {
    final Response response = await authenticateRequest(uri);
    expect(response.statusCode, equals(expectedResponseStatus));
  }

  void _expectResponse(Uri uri, void check(Response response)) {
    final authResultFuture = authenticateRequest(uri);
    authResultFuture.then((Response result) {
      check(result);
    });
    expect(authResultFuture, completes);
  }

  void runAuthTests(Uri uriNoJwt, Uri uriInvalidJwt, Uri uriWithValidJwt) {
    group('[no jwt]', () {
      test('response is 401 when no jwt in request', () {
        _expect(uriNoJwt, 401);
      });
    });

    group('[auth failure]', () {
      test('response is 401 when invalid jwt in request', () {
        _expect(uriInvalidJwt, 401);
      });
    });

    group('[auth success]', () {
      test('calls provided handler when valid jwt in request', () {
        _expectResponse(uriWithValidJwt, (_) {
          verify(handler.call(any)).called(1);
        });
      });

      test(
          'calls provided handler with session token in request',
          () {
            _expectResponse(uriWithValidJwt, (_) {
              verify(handler.call(argThat(new HasHeaderValue(
                  AC_DART_SESSION_PARAM_TOKEN, isNotNull)))).called(1);
            });
          },
          skip: 'no longer populating these. Are they worth bringing back?');

      test(
          'calls provided handler with hostkey in request',
          () {
            _expectResponse(uriWithValidJwt, (_) {
              verify(handler.call(argThat(new HasHeaderValue(
                  AC_DART_SESSION_PARAM_HOST_KEY, equals(hostKey))))).called(1);
            });
          },
          skip: 'no longer populating these. Are they worth bringing back?');

      test(
          'calls provided handler with username in request',
          () {
            _expectResponse(uriWithValidJwt, (_) {
              verify(handler.call(argThat(new HasHeaderValue(
                  AC_DART_SESSION_PARAM_USERNAME, equals(user))))).called(1);
            });
          },
          skip: 'no longer populating these. Are they worth bringing back?');

      test('returns provided handlers response', () {
        _expect(uriWithValidJwt, 200);
      });

      test(
          'response has token in header',
          () {
            _expectResponse(uriWithValidJwt, (response) {
              expect(response.headers,
                  containsPair('authorization', sessionToken));
            });
          },
          skip: 'TODO: may make sense to rework this test file with whole running acdart - no mocks');
    });
  }

  group('[host product authenticator]', () {
    final uriStrNoJwt =
        "http://localhost/hello-world?lic=none&tz=Australia%2FSydney&cp=%2Fjira&user_key=&loc=en-US&user_id=&xdm_e=http%3A%2F%2Fstorm%3A2990&xdm_c=channel-servlet-hello-world&xdm_p=1";
    final uriNoJwt = Uri.parse(uriStrNoJwt);

    final String invalidJwtStr =
        r'eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJleHAiOjEzOTQxNjYzNjcsInN1YiI6ImFkbWluIiwiaXNzIjoiamlyYTphZTM5MGQyOS0zMWIyLTRjMTItYTcxOS05ZGY2NGUzZTkyYjciLCJxc2giOiJlYjRlYzVmMDk4MGQwMWRhMzA2ZGI4OWZhNTdkZDE2MzU3NDY2NzQzNmRkNzIwZDVkOWM1Mjc5YzViNDVmN2E4IiwiaWF0IjoxMzk0MTY2MTg3fQ.bR8Z0MIguOg6xgHiav0quun8kTqXzNUKMKym-PFjZvc';
    final uriInvalidJwt = Uri.parse(
        "http://localhost/hello-world?lic=none&tz=Australia%2FSydney&cp=%2Fjira&user_key=&loc=en-US&user_id=&xdm_e=http%3A%2F%2Fstorm%3A2990&xdm_c=channel-servlet-hello-world&xdm_p=1&jwt=${invalidJwtStr}");

    String jwtToken = createProductHostToken(
        uriNoJwt.path, 'GET', uriNoJwt.queryParameters, sharedSecret, hostKey);

    final uriWithValidJwt = Uri.parse('$uriStrNoJwt&jwt=$jwtToken');

    setUp(() {
      authenticator = (new MiddlewareBuilder()
          .addMiddleware(exceptionHandler())
          .addMiddleware(authenticate([new JwtAuthenticator()],
              allowAnonymousAccess: false, allowHttp: true))).build();

      when(hostService.authenticateHostRequest(
              any, argThat(equals(uriWithValidJwt)), any)).thenReturn(
          new Future.value(new Right(new AcPrincipal(user, hostKey))));

      when(hostService.authenticateHostRequest(
              any, argThat(equals(uriInvalidJwt)), any)).thenReturn(
          new Future.value(
              new Left(
                  new AuthenticationFailure(
                      401,
                      new ConstraintViolations(
                          'oops',
                          new Set()
                            ..add(new ConstraintViolation('bugger')))))));
    });

    runAuthTests(uriNoJwt, uriInvalidJwt, uriWithValidJwt);
  });

  group(
      '[session token authenticator]',
      () {
        const String addonSecret = 'el secreto';
        const String addonKey = 'la addona';

        final uriStrNoJwt =
            "http://localhost/hello-world?lic=none&tz=Australia%2FSydney&cp=%2Fjira&" +
                "user_key=&loc=en-US&user_id=&xdm_e=http%3A%2F%2Fstorm%3A2990&" +
                "xdm_c=channel-servlet-hello-world&xdm_p=1";
        final uriNoJwt = Uri.parse(uriStrNoJwt);

        final String invalidJwtStr = r'eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9' +
            '.eyJleHAiOjEzOTQxNjYzNjcsInN1YiI6ImFkbWluIiwiaXNzIjoiamlyYTphZTM5MGQy' +
            'OS0zMWIyLTRjMTItYTcxOS05ZGY2NGUzZTkyYjciLCJxc2giOiJlYjRlYzVmMDk4MG' +
            'QwMWRhMzA2ZGI4OWZhNTdkZDE2MzU3NDY2NzQzNmRkNzIwZDVkOWM1Mjc5YzViNDVm' +
            'N2E4IiwiaWF0IjoxMzk0MTY2MTg3fQ.bR8Z0MIguOg6xgHiav0quun8kTqXzNUKMKym-PFjZvc';
        final uriInvalidJwt = Uri.parse(
            "http://localhost/hello-world?lic=none&tz=Australia%2FSydney&cp=%2Fjira&user_key=&" +
                "loc=en-US&user_id=&xdm_e=http%3A%2F%2Fstorm%3A2990&" +
                "xdm_c=channel-servlet-hello-world&xdm_p=1&jwt=${invalidJwtStr}");

        final String validJwtStr =
            createAddonSessionToken(addonSecret, addonKey, hostKey, user);

        final uriWithValidJwt = Uri.parse('$uriStrNoJwt&acpt=$validJwtStr');

        setUp(() {
          authenticator = (new MiddlewareBuilder()
              .addMiddleware(exceptionHandler())
              .addMiddleware(sessionTokenAuthenticator())).build();
//      authenticator = new SessionTokenAuthenticator(hostService);

          when(hostService.authenticateAddonSessionToken(validJwtStr))
              .thenReturn(
                  new Future.value(new Right(new AcPrincipal(user, hostKey))));

          when(hostService.authenticateAddonSessionToken(invalidJwtStr))
              .thenReturn(new Future.value(new Left(new AuthenticationFailure(
                  401,
                  new ConstraintViolations('oops',
                      new Set()..add(new ConstraintViolation('bugger')))))));
        });

        runAuthTests(uriNoJwt, uriInvalidJwt, uriWithValidJwt);
      },
      skip: 'hmmm should really have this, but it is now largely using shelf_auth session stuff');
}

class HasHeaderValue extends CustomMatcher {
  final String headerName;
  HasHeaderValue(String headerName, Matcher matcher)
      : this.headerName = headerName,
        super(
            "Request with header whose key is '$headerName' and value that is",
            "headers",
            matcher);

  String featureValueOf(Request actual) => actual.headers[headerName];
}
